from inspect    import getfullargspec
from enum import Enum

class AnswerSet:
    def __init__(self, dictionary={}):
        self.dictionary = dictionary

    @classmethod
    def parse(cls, a, interesting):
        if a.startswith('{') and a.endswith('}'):
            a = a[1:-1]
        a = a.split(', ')
        result = {}

        for predicate in interesting:
            result[predicate] = {}

        for e in a:
            neg = e.startswith('-')
            lpar = e.find('(')
            if lpar < 0:
                terms = ()
                predicate = e[neg:]
            else:
                predicate = e[neg:lpar]
                terms = e[lpar + 1:-1].split(',')
                if predicate in interesting:
                    terms = list(map(int, terms))
                    mapped = []
                    i = 0
                    for f in interesting[predicate]:
                        if i == len(terms):
                            break
                        if f == None or f == int or issubclass(f, Enum):
                            n = 1
                            mapped.append(f(int(terms[i])))
                        else:
                            spec = getfullargspec(f)
                            n = len(spec.args)
                            if n > 1 and spec.args[0] == 'self':
                                n -= 1
                            mapped.append(f(*terms[i:i+n]))
                        i += n
                    terms = tuple(mapped)
                else:
                    terms = tuple(map(int, terms))

            result[predicate][terms] = (not neg)
        return cls(result)

    def pretty(self, selection):
        selection = list(selection)
        selection.sort()
        width = max(map(len, selection))
        notes = []
        for predicate in selection:
            result = predicate.ljust(width) + ' = '

            if predicate not in self.dictionary:
                result += '∅'
            else:
                result += '{' + ', '.join(map(lambda y: '\033[3' + str(1 + y[1]) + 'm(' + ','.join(map(str, y[0]))  + ')\033[0m', self.dictionary[predicate].items())) + '}'
            notes.append(result)
        return notes

    def tofun(self, name):
        xs = self.dictionary[name]
        def f(y):
            fy = [x[0][1] for x in xs.items() if x[1] and y == x[0][0]]
            if len(fy) > 1:
                raise ValueError("Not a function since {} yields results {}!".format(y, fy))
            elif len(fy) == 0:
                return None
            else:
                return fy[0]
        return f

    def tobfun(self, name):
        xs = self.dictionary[name]
        def f(y):
            fy = [x[1] for x in xs.items() if y == x[0][0]]
            if len(fy) > 1:
                raise ValueError("Not a function since {} yields results {}!".format(y, fy))
            elif len(fy) == 0:
                return None
            else:
                return fy[0]
        return f

    def __getitem__(self, index):
        return self.dictionary[index]



from os      import stat
from os.path import exists
from stat    import S_ISFIFO
from random  import randint, random

from ..common import *

PIT_PROBABILITY = 0.2

emoji = False

GDSPC = '═' * (2 if emoji else 1) * 3
GSPC = '─' * (2 if emoji else 1) * 3
SPC = ' ' * (2 if emoji else 1)
GOLD = '🏆' if emoji else 'G'
WUMPUS = '👹' if emoji else 'W'
PIT = '⚫' if emoji else 'P'
WUMPUS_DEAD = '✨' if emoji else 'x'
AGENT = '🤖' if emoji else 'A'
ARROW = '🏹' if emoji else 'A'

MAX_ACTIONS = 1000

class World():
    def __init__(self, size=4, wumpus=None, gold=None, pits=None, optimum=None):
        self.shouldPaint = exists('world') and S_ISFIFO(stat('world').st_mode)

        self.agentLocation = Location(1,1)
        self.agentOrientation = Orientation.RIGHT
        self.agentAlive = True
        self.agentHasArrow = True
        self.agentHasGold = False
        self.agentInCave = True
        self.wumpusAlive = True
        self.pits = []
        self.gold = None
        self.optimum = optimum
        self.wumpus = None
        self.numActions = 0
        self.percept = None

        self.worldSize = size
        self.wumpus = wumpus
        self.gold = gold
        self.pits = pits

        if self.wumpus == None:
            # Choose wumpus location (anywhere except [1,1])
            x, y = 1, 1
            while x == 1 and y == 1:
                x, y = randint(1, size), randint(1, size)

            self.wumpus = Location(x, y)

        if self.gold == None:
            # Choose gold location (anywhere except [1,1])
            x, y = 1, 1
            while x == 1 and y == 1:# or wumpus.isAt(x, y):
                x, y = randint(1, size), randint(1, size)

            self.gold = Location(x, y)

        if self.pits == None:
            # Choose pit locations (anywhere except [1,1]) and location of gold
            self.pits = []
            for x in range(1, size + 1):
                for y in range(1, size + 1):
                    if (x == 1) and (y == 1): # or gold.isAt(x, y) or wumpus.isAt(x, y):
                        continue
                    if random() >= PIT_PROBABILITY:
                        continue
                    self.pits.append(Location(x, y))

        self.percept = Percept(
            ((self.agentLocation.isAdjacent(self.wumpus)) or (self.agentLocation == self.wumpus)),
            any(map(lambda pit: pit.isAdjacent(self.agentLocation), self.pits)),
            ((not self.agentHasGold) and (self.agentLocation == self.gold)),
            False,
            False
        )

    @classmethod
    def readFrom(cls, fname):
        size, wumpus, gold, optimum, pits = 4, None, None, None, []
        with open(fname, 'r') as f:
            for ln in f:
                ln = ln.strip()
                if ln == '' or ln.startswith('#'):
                    continue
                ln = ln.split(' ')
                if ln[0] == 'size':
                    size = int(ln[1])
                elif ln[0] == 'wumpus':
                    wumpus = Location(int(ln[1]), int(ln[2]))
                elif ln[0] == 'gold':
                    gold = Location(int(ln[1]), int(ln[2]))
                elif ln[0] == 'pit':
                    pits.append(Location(int(ln[1]), int(ln[2])))
                elif ln[0] == 'optimum':
                    optimum = int(ln[1])

        # World must contain wumpus and gold!
        if wumpus == None or gold == None:
            return None

        return cls(
            size=size, wumpus=wumpus, gold=gold, pits=pits, optimum=optimum
        )

    def execute(self, action: Action):
        # We assume the agent is alive and in the cave (i.e., game not over)
        self.numActions += 1
        if action is Action.GOFORWARD:
            newLocation: Location = self.agentLocation.getAdjacent(self.agentOrientation, self.worldSize)
            if newLocation == None:
                self.percept = Percept(
                    self.percept.stench,
                    self.percept.breeze,
                    self.percept.glitter,
                    True,
                    False
                )
            else:
                self.agentLocation = newLocation
                self.percept = Percept(
                    ((self.agentLocation.isAdjacent(self.wumpus)) or (self.agentLocation == self.wumpus)),
                    any(map(lambda pit: pit.isAdjacent(self.agentLocation), self.pits)),
                    ((not self.agentHasGold) and (self.agentLocation == self.gold)),
                    False,
                    False
                )

                fallsIntoPit = any(map(lambda pit: pit == self.agentLocation, self.pits))
                eatenByWumpus = self.wumpusAlive and self.wumpus == self.agentLocation
                self.agentAlive = not (fallsIntoPit or eatenByWumpus)
        elif action in {Action.TURNLEFT, Action.TURNRIGHT}:
            self.agentOrientation = self.agentOrientation.turn(action)
            self.percept = Percept(
                self.percept.stench,
                self.percept.breeze,
                self.percept.glitter,
                False,
                False
            )
        elif action == Action.GRAB:
            if not self.agentHasGold and self.agentLocation == self.gold:
                self.agentHasGold = True
                self.percept = Percept(
                    self.percept.stench,
                    self.percept.breeze,
                    False,
                    False,
                    False
                )
        elif action == Action.SHOOT:
            if not self.agentHasArrow:
                return
            self.agentHasArrow = False
            if not self.wumpusAlive:
                return

            if (((self.agentOrientation == Orientation.RIGHT) and
                 (self.agentLocation.x < self.wumpus.x) and
                 (self.agentLocation.y == self.wumpus.y)) or
                ((self.agentOrientation == Orientation.UP) and
                 (self.agentLocation.x == self.wumpus.x) and
                 (self.agentLocation.y < self.wumpus.y)) or
                ((self.agentOrientation == Orientation.LEFT) and
                 (self.agentLocation.x > self.wumpus.x) and
                 (self.agentLocation.y == self.wumpus.y)) or
                ((self.agentOrientation == Orientation.DOWN) and
                 (self.agentLocation.x == self.wumpus.x) and
                 (self.agentLocation.y > self.wumpus.y))):
                self.wumpusAlive = False
                self.percept = Percept(
                    self.percept.stench,
                    self.percept.breeze,
                    self.percept.glitter,
                    False,
                    True
                )
        elif action == Action.CLIMB:
            if self.agentLocation == Location(1, 1):
                self.agentInCave = False
                self.percept = Percept(False, False, False, False, False)

        return not self.isGameOver()

    def isGameOver(self):
        return not self.agentInCave or not self.agentAlive or self.numActions >= MAX_ACTIONS

    def getScore(self):
        # -1 for each action
        score = -self.numActions

        # -10 for shooting the arrow (already lost 1 for the action)
        if not self.agentHasArrow:
            score = score - 9

        # +1000 for leaving the cave with the gold
        if self.agentHasGold and not self.agentInCave:
            score = score + 1000

        # -1000 for dying
        if not self.agentAlive:
            score = score - 1000

        return score

    def paint(self):
        if not self.shouldPaint:
            return

        notes = [
            "  Agent has "+ARROW+": " + ('\033[32mYes\033[0m' if self.agentHasArrow else '\033[31mNo\033[0m'),
            "  Agent has "+GOLD+": " + ('\033[32mYes\033[0m' if self.agentHasGold else '\033[31mNo\033[0m'),
            "  Score:      " + SPC + str(self.getScore())
        ]

        def projAgent(l):
            return str(self.agentOrientation) if l == self.agentLocation else None

        paint(self.worldSize, [
            (self.wumpus, 'W' if self.wumpusAlive else 'w'),
            (self.pits, 'P'),
            (projAgent,),
            (self.gold, 'G' if not self.agentHasGold else 'g'),
        ], 'world', notes)

    def writeTo(self, fname):
        with open(fname, 'w') as f:
            f.write('\n'.join(
                [
                    'size {}'.format(self.worldSize),
                    'wumpus {} {}'.format(self.wumpus.x, self.wumpus.y),
                    'gold {} {}'.format(self.gold.x, self.gold.y)
                ] + [
                    'pit {} {}'.format(pit.x, pit.y) for pit in self.pits
                ] + [
                    ''
                ]
            ))



from .action import Action
from .orientation import Orientation
from .location import Location
from .percept import Percept
from .grid import paint

__all__ = ['Action', 'Orientation', 'Location', 'Percept', 'paint']



from .orientation import Orientation

class Location():
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def isAdjacent(self, other):
        x1, y1 = self
        x2, y2 = other

        isBelow = ((x1 == x2) and (y1 == (y2 - 1)))
        isAbove = ((x1 == x2) and (y1 == (y2 + 1)))
        isLeft  = ((x1 == (x2 - 1)) and (y1 == y2))
        isRight = ((x1 == (x2 + 1)) and (y1 == y2))

        return isAbove or isBelow or isLeft or isRight

    def getAdjacent(self, orientation, n):
        if orientation == Orientation.UP:
            return Location(self.x, self.y + 1) if self.y < n else None
        elif orientation == Orientation.DOWN:
            return Location(self.x, self.y - 1) if self.y > 1 else None
        elif orientation == Orientation.LEFT:
            return Location(self.x - 1, self.y) if self.x > 1 else None
        elif orientation == Orientation.RIGHT:
            return Location(self.x + 1, self.y) if self.x < n else None

    def __str__(self):
        return '({}, {})'.format(self.x, self.y)

    def __len__(self):
        return 2

    def __getitem__(self, index):
        if index == 0:
            return self.x
        elif index == 1:
            return self.y
        raise IndexError('')

    def __eq__(self, other):
        if other == None:
            return False

        if type(other) != Location:
            raise ValueError('Cannot compare ' + str(type(other)) + ' to Location!')

        return self.x == other.x and self.y == other.y

    def __hash__(self):
        return hash(str(self))



from heapq     import heappush, heappop
from itertools import count

from ..common    import Action, Location, Orientation
from ..simulator import World

class PerfectAgent():
    def __init__(self, world: World):
        if world.gold in world.pits:
            self.plan = [Action.CLIMB]
            return

        source, target = (Location(1, 1), Orientation.RIGHT), None
        paths, dist, seen, c = {source: []}, {}, {source: 0}, count()
        fringe = [(0, next(c), source)]

        while fringe:
            (d, _, v) = heappop(fringe)
            if v in dist:
                continue

            dist[v] = d
            if v[0] == world.gold:
                target = v
                break

            l, o = v
            neighbours = [([a], (l, o.turn(a)), 1) for a in {Action.TURNLEFT, Action.TURNRIGHT}]

            adj = l.getAdjacent(o, world.worldSize)
            if adj != None and adj not in world.pits:
                shoot = adj == world.wumpus
                actions = [Action.SHOOT] if shoot else []
                actions += [Action.GOFORWARD]
                neighbours.append((actions, (adj, o), 1 if not shoot else 10))

            for actions, u, cost in neighbours:
                vu_dist = dist[v] + cost
                # Cut off paths that are too costly.
                if vu_dist > 500:
                    continue
                if u not in seen or vu_dist < seen[u]:
                    seen[u] = vu_dist
                    heappush(fringe, (vu_dist, next(c), u))
                    paths[u] = paths[v] + actions

        # Gold not reachable.
        if target == None:
                 self.plan = [Action.CLIMB]
                 return

        goThere = paths[target]
        pickUpThatShiny = [Action.GRAB]
        turnAround = [Action.TURNLEFT, Action.TURNLEFT]
        comeBack = list(map(lambda x: x.mirror(), filter(lambda x: x != Action.SHOOT, goThere[::-1])))
        climbOut = [Action.CLIMB]

        # This is to avoid a useless turn as the last move coming back.
        if comeBack[-1] == Action.TURNRIGHT:
            comeBack = comeBack[:-1]

        self.plan = goThere + pickUpThatShiny + turnAround + comeBack + climbOut

    def process(self, percept):
        return self.plan.pop(0)



from collections import deque
from functools   import reduce
from itertools   import product
from os.path     import exists, dirname, join
from os          import remove, stat
from stat        import S_ISFIFO
from subprocess  import PIPE, STDOUT, run
from tempfile    import mkstemp

import logging

from .mode       import Mode
from ..asp       import unlite, fact, atom, AnswerSet
from ..common    import Action, Orientation, Location, paint
from ..simulator import World
from ..util      import dlv

logger = logging.getLogger('asp-agent')

plain = {
    # Required to return the desired action to the game.
    'do': (Action,),

    # The next three are needed to implement the autopilot.
    'autopilot': (),
    'goal': (Location,),
    'safe': (Location,),

    # For diagnosing de facto inconsistencies.
    'bad': (int,),

    # This one is needed for painting.
    'size': (int,),
}

interesting = {
    # For debugging, you may add other predicates here.
    # Note that painters are added below.
    'mode': (Mode,),
}

painted = (
    {
        'now': Orientation,
        'h':   int,
    },
    {
        'frontier': 'F',
        'safe':     'S',
        'pit':      'P',
        'next':     'N',
        'goal':     'G',
        'wumpus':   'W',
    }
)

extract = dict(plain)
extract.update(interesting)

extract.update(dict([
    (predicate, (Location,target)) for predicate, target in painted[0].items()
]))

extract.update(dict([
    (predicate, (Location,)) for predicate in painted[1].keys()
]))

class ASPAgent():
    def __init__(self):
        self.dlv = dlv()
        self.actions = []
        self.shot = None
        self.grabbed = None

        # Initially we do not know about stench/breeze/glitter anywhere.
        self.world = {}
        self.position = Location(1, 1)
        self.orientation = Orientation.RIGHT
        self.killed = False
        self.bumped = None
        self.previousAction = None

        # Assume some large world. Will get adjusted once we bump.
        self.size = 0xdeadbeef

        _, self.prog = mkstemp()
        unlite(join(dirname(__file__), 'agent.md'), self.prog)

        # To look at the ASP-Core compliant version, uncomment this.
        #unlite(join(dirname(__file__), 'agent.md'), 'agent.asp')

        self.paint = exists('agent') and S_ISFIFO(stat('agent').st_mode)

    def facts(self):
        # now/3 and killed/0 are certain.
        result = [
            fact(True, 'now', [*self.position, int(self.orientation)]),
            fact(self.killed, 'killed'),
        ]

        if self.bumped != None:
            result.append(fact(True, 'bumped', self.bumped))

        if self.shot != None:
            result.append(fact(True, 'shot', [*self.shot[0], int(self.shot[1])]))

        if self.grabbed != None:
            result.append(fact(True, 'grabbed', self.grabbed))

        for k, v in self.world.items():
            result += [
                fact(v[i], name, k) for i, name in enumerate(['stench', 'breeze', 'glitter'])
            ] + [
                fact(True, 'explored', k)
            ]

        return result

    def solve(self):
        return run(
            [
                self.dlv,
                '-n=1',
                '-silent',
                '-filter=' + ','.join(extract.keys()),
                self.prog,
                '--'
            ],
            stderr=STDOUT,
            stdout=PIPE,
            input='\n'.join(self.facts()),
            encoding='utf-8'
        )

    def bfs(self, safe, target):
        source = (self.position, self.orientation)
        visited = {source: None}
        queue = deque([source])
        while queue:
            node = queue.popleft()
            if node == target:
                path = []
                while node is not None:
                    v = visited[node]
                    if v == None:
                        break
                    act, node = v
                    path.append(act)
                return path[::-1]

            (l, o) = node
            neighbours = [(a, (l, o.turn(a))) for a in {Action.TURNLEFT, Action.TURNRIGHT}]

            adj = l.getAdjacent(o, self.size)
            if adj != None and (safe.get((adj,), False) or adj == target[0]):
                neighbours.append((Action.GOFORWARD, (adj, o)))

            for act, neighbour in neighbours:
                if neighbour not in visited:
                    visited[neighbour] = (act, node)
                    queue.append(neighbour)

        return None

    def process(self, percept):
        if percept.scream:
            self.killed = True

        if percept.bump:
            self.size = max(self.position.x, self.position.y)
            if self.bumped != None:
                logger.debug('We appear to be bumping a second time.')
            self.bumped = self.position.getAdjacent(self.orientation, self.size + 1)
        elif self.previousAction == Action.GOFORWARD:
            self.position = self.position.getAdjacent(self.orientation, self.size)
        elif self.previousAction in {Action.TURNLEFT, Action.TURNRIGHT}:
            self.orientation = self.orientation.turn(self.previousAction)
        elif self.previousAction == Action.SHOOT:
            if self.shot != None:
                logger.debug('We appear to be shooting a second time.')
            self.shot = (self.position, self.orientation)
        elif self.previousAction == Action.GRAB:
            self.grabbed = self.position

        if self.actions != []:
            self.previousAction = self.actions.pop(0)
            return self.previousAction

        self.world[self.position] = (percept.stench, percept.breeze, percept.glitter)

        result = self.solve().stdout

        if len(result) == 0:
            logger.debug('ASP program did not return any answer sets! Inconsistency?')
            logger.debug(result)
            return None

        if result.startswith('Best model: {'):
            start, end = result.find('{'), result.find('}')
            result = result[start:end+1]

        if result[0] != '{':
            logger.error('ASP Errors:\n  \033[31m' + result.replace('\n', '\n  ') + '\033[0m')
            return None

        result = AnswerSet.parse(result.strip().split('\n')[0], extract)
        size = next(l for (l,), sign in result['size'].items() if sign)

        if self.paint:
            paint(
                size,
                [(result.tofun(pred),) for pred, _ in painted[0].items()] +
                [(result.tobfun(pred), symbol) for pred, symbol in painted[1].items()],
                'agent',
                result.pretty(interesting.keys())
            )

        bads = result['bad'].items()
        if len(bads) > 0:
            for (x,), sign in bads:
                prefix = '%{} '.format(x)
                found = False
                with open(self.prog) as f:
                    for ln in f:
                        if ln.startswith(prefix):
                            found = True
                            logger.debug('\033[31mCONSISTENCY ' + str(x) + ': ' + ln[len(prefix):].strip() + '\033[0m')
                    if not found:
                        logger.debug('No comment describing bad({}). Add a line starting with \'%{} \' followed by a description.'.format(x, x))
            return None

        if result['autopilot'][()]:
            goal = next(l for (l,), sign in result['goal'].items() if sign)
            self.actions = min([self.bfs(result['safe'], (goal, o)) for o in Orientation], key=len)

        if self.actions != []:
            self.previousAction = self.actions.pop(0)
        else:
            self.previousAction = next(a for a, sign in result['do'].items() if sign)[0]

        return self.previousAction



from sys     import argv, exit
from random  import seed
from os      import urandom
from os.path import join
from glob    import glob
from time    import time

import logging

from ..common import *
from ..simulator import World
from ..agent import *

def generate(worldSize, seedV, base):
    fname = join(base, 'world-{}-{}.txt'.format(worldSize, seedV))

    world = World(worldSize)
    world.writeTo(fname)
    agent = PerfectAgent(world)

    while world.execute(agent.process(world.percept)):
        ''

    with open(fname, 'a') as f:
        f.write('optimum {}\n'.format(world.getScore()))

def instantiate(agentName, world):
    if agentName == 'proxy':
        return ProxyAgent()
    elif agentName == 'perfect':
        return PerfectAgent(world)
    elif agentName == 'asp':
        return ASPAgent()

def play(world, agentName):
    world.writeTo('last-world.txt')
    agent = instantiate(agentName, world)

    while True:
        world.paint()

        percept = world.percept
        action = agent.process(percept)

        if action == None:
            return None

        if not world.execute(action):
            return world.getScore()

def benchmark(bglob, agentName):
    for instance in glob(bglob):
        wumpusWorld = World.readFrom(instance)
        optimum = '{:5}'.format(wumpusWorld.optimum) if wumpusWorld.optimum else '  ?  '
        numPits = '{:2}'.format(len(wumpusWorld.pits))
        start = time()
        result = play(
            wumpusWorld,
            agentName
        )
        end = time()
        result = '!' if result is None else '{:5}'.format(result)
        elapsed = '{:7.4f}'.format(end - start)
        print('\t'.join([instance, numPits, optimum, result, elapsed]))

worldSize = 4
seedV = None
worldFile = None
agentName = 'proxy'
generationMode = False
base = None
bench = None

for arg, val in zip(argv[1:], argv[2:]):
    if arg == "-size":
        worldSize = max(2, int(val))
    elif arg == "-seed":
        seedV = bytes.fromhex(val)
    elif arg == "-world":
        worldFile = val
    elif arg == "-agent":
        agentName = val
    elif arg == '-generate':
        base = val
    elif arg == '-benchmark':
        bench = val

if seedV != None:
    seed(seedV)
else:
    seedV = urandom(4)
    seed(seedV)

seedV = seedV.hex()

if bench != None:
    benchmark(bench, agentName)
elif base != None:
    generate(worldSize, seedV, base)
else:
    logging.basicConfig(level=logging.DEBUG, format='')

    world = None
    if worldFile != None:
        world = World.readFrom(worldFile)
    else:
        print("World " + seedV)
        world = World(worldSize)

    play(
        world,
        agentName
    )



class Percept:
    def __init__(self, stench, breeze, glitter, bump, scream):
        self.stench = stench
        self.breeze = breeze
        self.glitter = glitter
        self.bump = bump
        self.scream = scream

    def __str__(self):
        return "Percept{" + "stench=" + str(self.stench) + ", breeze=" + str(self.breeze) + ", glitter=" + str(self.glitter) + ", bump=" + str(self.bump) + ", scream=" + str(self.scream) + "}"



from enum import Enum

class Action(Enum):
    GOFORWARD = 0
    TURNLEFT  = 1
    TURNRIGHT = 2
    GRAB      = 3
    SHOOT     = 4
    CLIMB     = 5

    def mirror(self):
        if self == Action.TURNLEFT:
            return Action.TURNRIGHT
        elif self == Action.TURNRIGHT:
            return Action.TURNLEFT
        else:
            return self

    def __str__(self):
        if self == Action.GOFORWARD:
            return 'F'
        elif self == Action.TURNLEFT:
            return 'L'
        elif self == Action.TURNRIGHT:
            return 'R'
        elif self == Action.GRAB:
            return 'G'
        elif self == Action.SHOOT:
            return 'S'
        elif self == Action.CLIMB:
            return 'C'
        else:
            return '?'

    def __int__(self):
        if self == Action.GOFORWARD:
            return 0
        elif self == Action.TURNLEFT:
            return 1
        elif self == Action.TURNRIGHT:
            return 2
        elif self == Action.GRAB:
            return 3
        elif self == Action.SHOOT:
            return 4
        elif self == Action.CLIMB:
            return 5
        else:
            return None



from enum import Enum

class Mode(Enum):
    EXPLORE = 0
    ESCAPE  = 1
    KILL    = 2
    GRAB    = 3

    def __str__(self):
        if self == Mode.EXPLORE:
            return 'explore'
        elif self == Mode.ESCAPE:
            return 'escape'
        elif self == Mode.KILL:
            return 'kill'
        elif self == Mode.GRAB:
            return 'grab'
        else:
            return '?'

    def toSymbol(self):
        if self == Mode.EXPLORE:
            return 'explore'
        elif self == Mode.ESCAPE:
            return 'escape'
        elif self == Mode.KILL:
            return 'kill'
        elif self == Mode.GRAB:
            return 'grab'
        else:
            return '?'



from .lite import unlite
from .answer import AnswerSet

__all__ = ['unlite', 'atom', 'fact', 'AnswerSet']

def unlit(fname):
    with open(fname, 'r') as f:
        return [ln[4:-1] if ln.startswith('    ') else '%' + ln[:-1] for ln in f]

def atom(sign, predicate, terms=[]):
    result = ''
    if sign == None:
        result += 'not '
    elif sign == False:
        result += '-'
    elif sign != True:
        raise ValueError('Sign must be None or boolean!')
    result += predicate
    if len(terms ) == 0:
        return result
    return result + '(' + ','.join(map(str, terms)) + ')'

def fact(sign, predicate, terms=[]):
    return '' if sign == None else atom(sign, predicate, terms) + '.'



from os             import chmod, getcwd
from os.path        import isfile, join
from shutil         import which
from stat           import S_IRGRP, S_IRWXU, S_IWGRP
from sys            import platform, maxsize
from urllib.request import urlopen

def downloadDlv(fname):
    # Taken from http://www.dlvsystem.com/dlv/
    if platform.startswith('linux'):
        if maxsize == 2 ** 31 - 1:
            url = 'http://www.dlvsystem.com/files/dlv.i386-linux-elf-static.bin'
        else:
            url = 'http://www.dlvsystem.com/files/dlv.x86-64-linux-elf-static.bin'
    if platform.startswith('darwin'):
        url = 'http://www.dlvsystem.com/files/dlv.i386-apple-darwin.bin'
    if platform.startswith('cygwin') or platform.startswith('win'):
        url = 'http://www.dlvsystem.com/files/dlv.mingw.exe'

    res = urlopen(url)

    if res.status != 200:
        print('Failed to download DLV! GET {} resulted in HTTP {}'.format(url, res.status))
        return

    with open(fname, 'wb') as f:
        f.write(res.read())

def dlv():
    here = join(getcwd(), 'dlv')
    fname = which('dlv')

    if fname != None:
        return fname
    if isfile(here):
        return here
    else:
        print('NOTE: Could not find DLV! Will attempt to download it.')
        downloadDlv(here)
        chmod(here, S_IRGRP | S_IRWXU | S_IWGRP)
        return here



from math import ceil, sqrt

from .location import Location

right = ('╟')
up = ('╔', '╤')

# painters should be a list of tuples:
# (filter, t, f)
def paint(n, painters, fname, notes=[]):
    image = []
    height, width = 3, 6
    image.append('    ╔' + '╤'.join(['═' * width for z in range(n)]) + '╗')
    for y in list(range(1, n + 1))[::-1]:
        for i in range(height):
            ln = '  '
            if i == int(height / 2):
                ln += (str(y))
            else:
                ln += (' ')
            ln += (' ║')
            for x in range(1, n + 1):
                l = Location(x, y)
                for j in range(width):
                    idx = i * width + j
                    yp, np = None, None
                    if len(painters) <= idx:
                        ln += (' ')
                        continue
                    if len(painters[idx]) == 3:
                        p, yp, np = painters[idx]
                    elif len(painters[idx]) == 2:
                        p, yp = painters[idx]
                        np = ' '
                    elif len(painters[idx]) == 1:
                        p = painters[idx][0]
                        np = ' '
                    else:
                        raise ValueError('x')

                    if type(p) == dict:
                        if l in p:
                            p = False
                        elif (l,) in p:
                            p = p[(l,)]
                        else:
                            p = False
                    elif type(p) == list:
                        if l in p:
                            p = True
                        else:
                            p = False
                    elif type(p) == Location:
                        p = l == p
                        #p = False
                    elif callable(p):
                        p = p(l)
                        if p == None:
                            p = False
                        elif type(p) == bool:
                            if yp == None:
                                yp = str(p)
                        else:
                            yp = str(p)
                            p = True
                    else:
                        raise ValueError('Unknown painter!')

                    ln += (yp if p else np)
                    p = False
                if x < n:
                    ln += ('│')
                else:
                    ln += ('║')
            image.append(ln)
        if y > 1:
           image.append('    ╟' + '┼'.join(['─' * width for z in range(n)]) + '╢')

    image.append('    ╚' + '╧'.join(['═' * width for z in range(n)]) + '╝')
    image.append('     ' + ' '.join([str(z + 1).ljust(int(width - 1 / 2)).rjust(width) for z in range(n)]))

    imw = max(map(len, image))

    while len(notes) > len(image) - 2:
        image = ['    ' + ' ' * 30 + notes.pop(0)] + image

    for i, note in enumerate(notes):
        image[-(i + 2)] += '    ' + ' ' * (max(0, 30 - imw)) + note

    with open(fname, 'w') as f:
        f.write('\n' * 20)
        f.write('\n'.join(image))
        f.write('\n' * 3)



from ..common import Action

class ProxyAgent():
    def process(self, percept):
        while True:
            c = input('Action? {f,l,r,g,s,c} ')[0].lower()
            if c == 'f':
                return Action.GOFORWARD
            elif c == 'l':
                return Action.TURNLEFT
            elif c == 'r':
                return Action.TURNRIGHT
            elif c == 'g':
                return Action.GRAB
            elif c == 's':
                return Action.SHOOT
            elif c == 'c':
                return Action.CLIMB
            else:
                print('Huh?')



from enum import Enum

from .action import Action

class Orientation(Enum):
    RIGHT = 0
    UP    = 1
    LEFT  = 2
    DOWN  = 3

    def turn(self, action: Action):
        if action == Action.TURNLEFT:
            if self == Orientation.RIGHT:
                return Orientation.UP
            elif self == Orientation.UP:
                return Orientation.LEFT
            elif self == Orientation.LEFT:
                return Orientation.DOWN
            elif self == Orientation.DOWN:
                return Orientation.RIGHT
        elif action == Action.TURNRIGHT:
            if self == Orientation.RIGHT:
                return Orientation.DOWN
            elif self == Orientation.UP:
                return Orientation.RIGHT
            elif self == Orientation.LEFT:
                return Orientation.UP
            elif self == Orientation.DOWN:
                return Orientation.LEFT
        else:
            return None

    def __str__(self):
        if self == Orientation.RIGHT:
            return '→'#'R'
        elif self == Orientation.UP:
            return '↑'#'U'
        elif self == Orientation.LEFT:
            return '←'#'L'
        elif self == Orientation.DOWN:
            return '↓'#'D'
        else:
            return '?'

    def __int__(self):
        if self == Orientation.RIGHT:
            return 0
        elif self == Orientation.UP:
            return 1
        elif self == Orientation.LEFT:
            return 2
        elif self == Orientation.DOWN:
            return 3
        else:
            return None



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="hakuna_matata",
    version="0.0.3",
    author="Lorenz Leutgeb",
    author_email="lorenz@leutgeb.xyz",
    description="A logic-based agent that plays Hunt the Wumpus",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/lorenzleutgeb/ils",
    packages=setuptools.find_packages(),
    entry_points={'console_scripts': [
        'hakuna_matata = hakuna_matata.cli.__main__',
    ]},
    license='MIT',
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Education",
    ),
)



from functools import reduce
from sys       import stdin, stdout
from itertools import combinations

tab = '\t'

# Line feed.
lf = '\n'

cons = ':-'
wcons = ':~'

aggregates = {'#count', '#sum', '#times', '#min', '#max'}

tablen = 8

def keepSpaces(s):
    def f(acc, x):
        if x == '' and acc != []:
            acc[-1] += ' '
        else:
            acc.append(x)
        return acc

    return reduce(f, s.split(' '), [])

def atomize(s):
    s = keepSpaces(s)
    dneg = s[0] == 'not'
    if len(s) - dneg == 1:
        return ('not ' if dneg else '') + s[dneg]
    return ('not ' if dneg else '') + s[dneg] + '(' + ','.join(s[dneg + 1:]) + ')'

def ignore(ln):
    return len(ln) < 2 or ln[0] != tab

def isBody(ln):
    return len(ln) > 0 and ln[0] == tab

def unlite(src, dst):
    with open(src, 'r') as srcf:
        with open(dst, 'w') as dstf:
            _unlite(srcf, dstf)

def _unlite(src, dst):
    w = dst.write

    def unindent(previously, now):
        if previously == now and now == 0:
            return

        if previously < now:
            return

        if previously == now and now == 1:
            w('.' + lf)
            return

        final = now <= 1
        w('}' * (previously - now - final - 1) + ('.' if final else '') + lf)

    # Indicates that we are currently reading bodies.
    bodyMode = False
    prevIndentation = 0

    for ln in src:
        # If the first line is not a tab, or the line is empty, just pass it on.
        if ignore(ln):
            indentation = 0
            # if prevIndentation > indentation and prevIndentation > 2:
            #     w('}' * (prevIndentation - indentation))
            # prevIndentation = indentation
            # if bodyMode:
            #     w('.' + lf)
            unindent(prevIndentation, indentation)
            prevIndentation = indentation
            w('%' + ln)
            bodyMode = False
            continue

        if '"' in ln or "'" in ln:
            raise ValueError('String literals are not supported yet.')

        # Strip away the first tab.
        ln = ln[1:]
        readBody = isBody(ln)

        # We recognized a body, now strip away the next tab.
        if readBody:
            ln = ln[1:]

        if not readBody:
            ln = ln.lstrip()

            # Preprocessor rules (constants).
            if ln[0] == '#':
                w(ln)
                continue

            # Since we are in head mode we can strip away (no need to read any further indentation).
            ln = ln.lstrip()

            # We were in body mode, so terminate the previous body.
            # if bodyMode:
            #     w('.')
            unindent(prevIndentation, 1)
            prevIndentation = 1
            bodyMode = False

            # Check for a hard constraint.
            if len(ln) == 2 and ln[0] == cons[1]:
                w(cons + lf)
                continue

            # Check for a weak constraint.
            if len(ln) == 2 and ln[0] == wcons[1]:
                w(wcons + lf)
                continue

            # TODO: This breaks for string literals (there may be a pipe in  a string literal).
            ln = ln.split('|')

            # Special case: Facts over fixed integer ranges.
            # TODO: This breaks for string literals (there might be two dots inside a string literal).
            if len(ln) == 1 and '..' in ln[0]:
                w(atomize(ln[0].strip()) + '.' + lf)
                # Since this fact auto-terminates, we fake prevIndentation.
                prevIndentation = 0
                continue

            w(tab + ' v '.join(map(lambda x: atomize(x.strip()), ln)) + ' ' + cons + ' ')
            continue

        # We are treating bodies from here on.
        if not bodyMode:
            w(lf)

        indentation = 2
        while ln[0] == tab:
            indentation += 1
            ln = ln[1:]

        # Strip away trailing newline and other stuff...
        ln = ln.strip()

        isAggregate = False
        for aggregate in aggregates:
            if ln.startswith(aggregate):
                isAggregate = True

        # We first handle the easier case of non-aggregates.
        if not isAggregate:
            unindent(prevIndentation, indentation)

            if bodyMode and prevIndentation >= indentation:
                w(',')

            w(tab * indentation + atomize(ln))
            prevIndentation = indentation
            bodyMode = True
            continue

        ln = keepSpaces(ln)
        if bodyMode:
            w(lf + ',')

        w(tab * indentation + ln[1] + ' = ' + ln[0] + '{' + (','.join(ln[2:]) if len(ln) > 3 else ln[2]) + ': ' + lf)
        bodyMode = True

    if bodyMode:
        w('.')
